from pymongo import MongoClient
from interface import implements
from typing import List
from liteflow.core.abstractions import IPersistenceProvider
from liteflow.core.models import WorkflowInstance, EventSubscription, Event
from .converters import *


class MongoPersistenceProvider(implements(IPersistenceProvider)):

    def __init__(self, connection_string, database):
        self._client = MongoClient(connection_string)
        self._database = self._client[database]
        self._workflow_collection = self._database['workflows']
        self._subscription_collection = self._database['subscriptions']
        self._event_collection = self._database['events']

    def create_workflow(self, workflow: WorkflowInstance):
        data = dump_workflow_instance(workflow)
        workflow.id = str(self._workflow_collection.insert_one(data).inserted_id)
        return workflow.id

    def persist_workflow(self, workflow: WorkflowInstance):
        data = dump_workflow_instance(workflow)
        self._workflow_collection.replace_one({ '_id': ObjectId(workflow.id) }, data)

    def get_workflow_instance(self, id) -> WorkflowInstance:
        data = self._workflow_collection.find_one({ "_id": ObjectId(id)})
        return load_workflow_instance(data)

    def get_runnable_instances(self) -> []:
        result = []
        for item in self._workflow_collection.find(filter={"status": WorkflowInstance.RUNNABLE}, projection={"_id": True}):
            result.append(item["_id"])
        return result

    def create_subscription(self, subscription: EventSubscription):
        data = dump_subscription(subscription)
        subscription.id = str(self._subscription_collection.insert_one(data).inserted_id)
        return subscription.id

    def get_subscriptions(self, event_name, event_key, effective_date) -> []:
        result = []
        for item in self._subscription_collection.find(filter={
            "event_name": event_name,
            "event_key": event_key,
            "subscribe_as_of": {"$lte": effective_date}
        }):
            result.append(load_subscription(item))
        return result

    def terminate_subscription(self, subscription_id):
        self._subscription_collection.delete_one({"_id": subscription_id})

    def create_event(self, evt: Event):
        data = dump_event(evt)
        evt.id = str(self._event_collection.insert_one(data).inserted_id)
        return evt.id

    def get_event(self, event_id) -> Event:
        data = self._event_collection.find_one({"_id": ObjectId(event_id)})
        return load_event(data)

    def mark_event_processed(self, event_id):
        self._event_collection.update_one({"_id": event_id}, {"$set": {"is_processed": True}})

    def mark_event_unprocessed(self, event_id):
        self._event_collection.update_one({"_id": event_id}, {"$set": {"is_processed": False}})

    def get_runnable_events(self, effective_date) -> []:
        result = []
        for item in self._event_collection.find(filter={
            "is_processed": False,
            "event_time": {"$lte": effective_date}
        }, projection={"_id": True}):
            result.append(item["_id"])
        return result

    def get_events(self, event_name, event_key, effective_date) -> []:
        result = []
        for item in self._event_collection.find(filter={
            "event_name": event_name,
            "event_key": event_key,
            "event_time": {"$gte": effective_date}
        }, projection={"_id": True}):
            result.append(item["_id"])
        return result

    def persist_errors(self, errors: []):
        pass




import pickle
from typing import Dict
from bson.objectid import ObjectId
from liteflow.core.models import WorkflowInstance, EventSubscription, Event, ExecutionPointer


def dump_execution_pointer(source: ExecutionPointer):
    return {
        'id': source.id,
        'step_id': source.step_id,
        'active': source.active,
        'sleep_until': source.sleep_until,
        'persistence_data': pickle.dumps(source.persistence_data),
        'start_time': source.start_time,
        'end_time': source.end_time,
        'event_name': source.event_name,
        'event_key': source.event_key,
        'event_published': source.event_published,
        'event_data': pickle.dumps(source.event_data),
        'context_item': pickle.dumps(source.context_item),
        'predecessor_id': source.predecessor_id,
        'outcome': source.outcome,
        'status': source.status,
        'retry_count': source.retry_count,
        'children': pickle.dumps(source.children),
        'scope': pickle.dumps(source.scope)
    }


def dump_workflow_instance(source: WorkflowInstance):
    result = {
        'workflow_definition_id': source.workflow_definition_id,
        'version': source.version,
        'description': source.description,
        'next_execution': source.next_execution,
        'status': source.status,
        'data': pickle.dumps(source.data),
        'create_time': source.create_time,
        'complete_time': source.complete_time,
        'execution_pointers': [dump_execution_pointer(x) for x in source.execution_pointers]
    }

    if source.id is not None:
        result['_id'] = ObjectId(source.id)

    return result


def dump_subscription(source: EventSubscription):
    result = {
        'event_name': source.event_name,
        'event_key': source.event_key,
        'step_id': source.step_id,
        'workflow_id': source.workflow_id,
        'subscribe_as_of': source.subscribe_as_of
    }

    if source.id is not None:
        result['_id'] = ObjectId(source.id)

    return result


def dump_event(source: Event):
    result = {
        'event_name': source.event_name,
        'event_key': source.event_key,
        'event_data': pickle.dumps(source.event_data),
        'event_time': source.event_time,
        'is_processed': source.is_processed
    }

    if source.id is not None:
        result['_id'] = ObjectId(source.id)

    return result


def load_execution_pointer(source: Dict) -> ExecutionPointer:
    result = ExecutionPointer()
    result.id = source['id']
    result.step_id = source['step_id']
    result.active = source['active']
    result.sleep_until = source['sleep_until']
    result.persistence_data = pickle.loads(source['persistence_data'])
    result.start_time = source['start_time']
    result.end_time = source['end_time']
    result.event_name = source['event_name']
    result.event_key = source['event_key']
    result.event_published = source['event_published']
    result.event_data = pickle.loads(source['event_data'])
    result.context_item = pickle.loads(source['context_item'])
    result.predecessor_id = source['predecessor_id']
    result.outcome = source['outcome']
    result.status = source['status']
    result.retry_count = source['retry_count']
    result.children = pickle.loads(source['children'])
    result.scope = pickle.loads(source['scope'])

    return result


def load_workflow_instance(source: Dict) -> WorkflowInstance:
    result = WorkflowInstance()
    result.id = str(source['_id'])
    result.workflow_definition_id = source['workflow_definition_id']
    result.version = source['version']
    result.description = source['description']
    result.data = pickle.loads(source['data'])
    result.next_execution = source['next_execution']
    result.status = source['status']
    result.create_time = source['create_time']
    result.complete_time = source['complete_time']
    result.execution_pointers = [load_execution_pointer(x) for x in source['execution_pointers']]

    return result


def load_subscription(source: Dict) -> EventSubscription:
    result = EventSubscription()
    result.id = str(source['_id'])
    result.event_name = source['event_name']
    result.event_key = source['event_key']
    result.step_id = source['step_id']
    result.workflow_id = source['workflow_id']
    result.subscribe_as_of = source['subscribe_as_of']

    return result


def load_event(source: Dict) -> Event:
    result = Event()
    result.id = str(source['_id'])
    result.event_name = source['event_name']
    result.event_key = source['event_key']
    result.event_data = pickle.loads(source['event_data'])
    result.event_time = source['event_time']
    result.is_processed = source['is_processed']

    return result



from setuptools import setup

with open("readme.md", "r") as fh:
    long_description = fh.read()

setup(
    name='liteflow.providers.mongo',
    version='0.3.1',
    description='MongoDB persistence provider for LiteFlow',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Daniel Gerlag',
    author_email='daniel@gerlag.ca',
    license='MIT',
    namespace_packages=['liteflow'],
    packages=['liteflow.providers.mongo'],
    zip_safe=False,
    install_requires=[
        'liteflow.core>=0.3',
        'pymongo>=3.6.1',
        'python-interface>=1.4.0'
    ],
    url="https://github.com/danielgerlag/liteflow",
    python_requires='>=3.6',
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    )
)